def shift_arr_by_one(arr):
    """
    Shifts an array to the right by 1 element
    e.g. shift_arr([1, 2, 3]) -> [3, 1, 2]
    """
    return arr[-1:] + arr[:-1]


def is_circular_match(list_1, list_2):
    """Does any rotation of the elements in `list_1` match `list_2`?"""

    str_1 = " ".join(map(str, list_1))
    str_2 = " ".join(map(str, list_2))

    if len(str_1) != len(str_2):
        return False

    str_2_expanded = str_2 + " " + str_2
    return str_1 in str_2_expanded


# def is_partial_circular_match(list_1, list_2):
#     """Is any rotation of the elements in `list_1` contained within `list_2`?"""

#     # LEFT OFF:
#     # is ["F1", "A#0", "F#0"] a subset of an Indian-Japan Pentatonic?
#     # how about ["F0", "F#0", "A#0", "C#1"] and ["", "", "", "C#1"]

#     # may be able to adapt the `slots` logic from `select_voicing()` here

#     one_indices_1 = [ix for ix, i in enumerate(list_1) if i == 1]
#     one_indices_2 = [ix for ix, i in enumerate(list_2) if i == 1]

#     # enforce that a note in each slot gets selected
#     choices = {}

#     while len(choices) < sum(rotation):

#         choice = np.random.choice(one_indices)

#         slot = choice % 12

#         if slot not in choices:
#             choices[slot] = choice



from chord_progressions.type_templates import (
    TYPE_TEMPLATES,
    get_template_from_midi_nums,
    get_template_from_notes,
    get_template_from_pitch_classes,
    get_template_from_template_str,
    get_type_from_midi_nums,
    get_type_from_notes,
    get_type_from_type_num,
    get_type_num_from_type,
    get_types_from_type_num_str,
    notes_match_chord_type,
)


def test_type_templates_format():
    assert type(TYPE_TEMPLATES) == dict


def test_get_template_from_pitch_classes():
    assert get_template_from_pitch_classes([0, 4, 7]) == [
        1,
        0,
        0,
        0,
        1,
        0,
        0,
        1,
        0,
        0,
        0,
        0,
    ]


def test_get_template_from_midi_nums():
    assert get_template_from_midi_nums([48, 60]) == [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]


def test_get_template_from_template_str():
    template = [1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0]
    assert get_template_from_template_str("101010101010") == template


def test_get_type_from_type_num():
    assert get_type_from_type_num(63) == "major chord"


def test_get_types_from_type_num_str():
    assert get_types_from_type_num_str("59_63") == ["minor chord", "major chord"]


def test_get_type_num_from_type():
    assert get_type_num_from_type("major chord") == 63


def test_notes_match_chord_type():
    assert notes_match_chord_type(["C4", "E4", "G4"], "major chord")


def test_get_type_from_notes():
    assert get_type_from_notes(["C4", "C3"]) == "unison"


def test_get_type_from_midi_nums():
    assert get_type_from_midi_nums([48, 60]) == "unison"


def test_get_template_from_notes():
    expectations = [
        {
            "notes": ["C-1", "A4", "G9"],
            "template": [1, 0, 0, 0, 0, 0, 0, 1, 0, 1, 0, 0],
        },
        {"notes": ["C4", "E4", "G4"], "template": [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0]},
    ]
    for ex in expectations:
        assert get_template_from_notes(ex["notes"]) == ex["template"]



from chord_progressions import logger
from chord_progressions.pitch import (
    get_note_from_midi_num,
    get_pitch_class_from_midi_num,
    get_pitch_class_from_note,
)
from chord_progressions.utils import is_circular_match

"""
From https://web.archive.org/web/20221116234056/https://vladimir_ladma.sweb.cz/english/music/structs/mus_rot.htm

Further readings:

Forte, Allen. The Structure of Atonal Music. New Haven: Yale, 1973

Solomon, Larry. "The List of Chords, Their Properties, and Uses", Interface, Journal of New Music Research, Vol. 11 (1982)

Strauss, Joseph N. Post-Tonal Theory, Prentice-Hall, 1990.

Ideally we'd like to reduce this to a smaller set, say by remapping.
I wonder if there is a perceptual study that can be done for this.

remap = {
 'minor-ninth chord': 'minor-seventh chord',
 'major-ninth chord': 'major-seventh chord',
 'diminished chord': 'tritone',
 'dominant-ninth,major-minor|prometheus pentamirror': 'tritone',
 'italian sixth|incomplete dominant-seventh chord 1': 'tritone',
 'dominanth-11th|natural / genuine / lydian hexachord': 'tritone',
 'pyramid': 'tritone',
}
"""

TYPE_TEMPLATES = {
    "": "",
    "unison": "100000000000",
    "semitone": "110000000000",
    "whole-tone": "101000000000",
    "bach / chromatic trimirror": "111000000000",
    "minor third": "100100000000",
    "phrygian trichord": "110100000000",
    "minor trichord": "101100000000",
    "major third": "100010000000",
    "major-minor trichord 1": "110010000000",
    "whole-tone trichord": "101010000000",
    "major-second tetracluster 2": "111010000000",
    "major-minor trichord 2": "100110000000",
    "alternating tetramirror": "110110000000",
    "major-second tetracluster 1": "101110000000",
    "chromatic pentamirror": "111110000000",
    "perfect fourth": "100001000000",
    "incomplete minor-seventh chord": "101001000000",
    "minor-third tetracluster 2": "111001000000",
    "incomplete dominant-seventh chord 2": "100101000000",
    "dorian tetrachord|phrygian tetrachord": "110101000000",
    "phrygian tetrachord|minor tetramirror": "101101000000",
    "major-second pentacluster 2": "111101000000",
    "chromatic mezotetrachord|arabian tetramirror": "110011000000",
    "lydian tetrachord|major tetrachord": "101011000000",
    "minor-second major pentachord": "111011000000",
    "minor-third tetracluster 1": "100111000000",
    "spanish pentacluster": "110111000000",
    "major-second pentacluster 1": "101111000000",
    "chromatic hexamirror": "111111000000",
    "tritone": "100000100000",
    "rite chord 2|tritone-fourth 1": "110000100000",
    "italian sixth|incomplete dominant-seventh chord 1": "101000100000",
    "major-third tetracluster 2": "111000100000",
    "diminished chord": "100100100000",
    "minor-second diminished tetrachord": "110100100000",
    "harmonic-minor tetrachord": "101100100000",
    "blues pentacluster": "111100100000",
    "incomplete half-dim-seventh chord": "100010100000",
    "all-interval tetrachord 1": "110010100000",
    "whole-tone tetramirror": "101010100000",
    "tritone-expanding pentachord": "111010100000",
    "major-third diminished tetrachord": "100110100000",
    "alternating pentachord 1": "110110100000",
    "tritone-symmetric pentamirror": "101110100000",
    "rite chord 1|tritone-fourth 2": "100001100000",
    "double fourth tetramirror": "110001100000",
    "all-interval tetrachord 2": "101001100000",
    "oriental pentacluster 1": "111001100000",
    "perfect-fourth diminished tetrachord": "100101100000",
    "locrian pentamirror": "110101100000",
    "alternating pentachord 2": "101101100000",
    "major-third tetracluster 1": "100011100000",
    "oriental pentacluster 2": "110011100000",
    "tritone-contracting pentachord": "101011100000",
    "minor-third pentacluster": "100111100000",
    "chromatic heptamirror": "111111100000",
    "quartal trichord": "101000010000",
    "perfect fourth tetramirror": "111000010000",
    "minor chord": "100100010000",
    "all-interval tetrachord 3": "110100010000",
    "major-second minor tetrachord": "101100010000",
    "major-third pentacluster 1": "111100010000",
    "major chord": "100010010000",
    "major-diminished tetrachord": "110010010000",
    "major-second major tetrachord": "101010010000",
    "major-seventh pentacluster 2": "111010010000",
    "major-minor tetramirror": "100110010000",
    "major-minor-dim pentachord 1": "110110010000",
    "center-cluster pentachord 1": "101110010000",
    "minor-second quartal tetrachord": "110001010000",
    "quartal tetramirror": "101001010000",
    "double-seconds triple-fourth pentachord 1": "111001010000",
    "perfect-fourth minor tetrachord": "100101010000",
    "phrygian pentachord": "110101010000",
    "dorian / minor pentachord": "101101010000",
    "perfect-fourth major tetrachord": "100011010000",
    "gypsy / semiditonic pentachord 1": "110011010000",
    "major / ionic pentachord": "101011010000",
    "center-cluster pentachord 2": "100111010000",
    "messiaen's truncated 5|lendvai's|double tritone tetramirror": "110000110000",
    "tritone quartal tetrachord": "101000110000",
    "double pentacluster1": "111000110000",
    "minor-diminished tetrachord": "100100110000",
    "javanese pentachord": "110100110000",
    "gypsy / semiditonic pentachord 2": "101100110000",
    "all-interval tetrachord 4": "100010110000",
    "balinese pentachord": "110010110000",
    "lydian pentachord": "101010110000",
    "major-minor-dim pentachord 2": "100110110000",
    "alternating hexamirror": "110110110000",
    "double pentacluster 2": "110001110000",
    "double-seconds triple-fourth pentachord 2": "101001110000",
    "double-cluster hexamirror": "111001110000",
    "major-seventh pentacluster 1": "100101110000",
    "major-third pentacluster 2": "100011110000",
    "chromatic octamirror": "111111110000",
    "augmented chord": "100010001000",
    "minor-augmented tetrachord": "110010001000",
    "augmented pentacluster 1": "111010001000",
    "augmented-major tetrachord": "100110001000",
    "minor-major ninth chord": "110110001000",
    "augmented pentacluster 2": "101110001000",
    "major-seventh chord": "110001001000",
    "half-diminished seventh chord": "101001001000",
    "diminished pentacluster 1": "111001001000",
    "minor-seventh chord": "100101001000",
    "major-ninth chord": "110101001000",
    "diminished-major ninth chord": "101101001000",
    "syrian pentatonic|major-augmented ninth chord": "110011001000",
    "diminished-augmented ninth chord": "101011001000",
    "center-cluster pentamirror": "100111001000",
    "french-sixth chord|messiaen's truncated 6": "101000101000",
    "bardos's|assymetric pentamirror": "111000101000",
    "dominant-seventh / german-sixth chord": "100100101000",
    "kumoi pentachord 2": "110100101000",
    "augmented-sixth pentachord 1": "101100101000",
    "enigmatic pentachord 1": "110010101000",
    "whole-tone pentamirror": "101010101000",
    "augmented-diminished ninth chord": "100110101000",
    "balinese pelog pentatonic|korean": "110001101000",
    "augmented-sixth pentachord 2|indian hindola|javan pentatonic": "101001101000",
    "minor-diminished ninth chord": "100101101000",
    "locrian hexachord|suddha saveriraga": "110101101000",
    "super-locrian hexamirror": "101101101000",
    "indian-japan pentatonic": "101100011000",
    "persian pentamirror": "110010011000",
    "enigmatic pentachord 2|altered pentatonic": "101010011000",
    "lebanese pentachord|augmented-minor chord": "100110011000",
    "megha or 'cloud' raga": "101110011000",
    "korean|kumoi pentachord 1": "101001011000",
    "minor-ninth chord": "100101011000",
    "phrygian hexamirror": "110101011000",
    "minor hexachord": "101101011000",
    "gypsy hexatonic": "110011011000",
    "melodic-minor hexachord": "101011011000",
    "messiaen's 5": "111000111000",
    "diminished pentacluster 2": "100100111000",
    "chromatic nonamirror": "111111111000",
    "diminished-seventh chord": "100100100100",
    "diminished minor-ninth chord": "110100100100",
    "flat-ninth pentachord|ranjaniraga": "101100100100",
    "neapolitan pentachord 1": "110010100100",
    "dominant-ninth,major-minor|prometheus pentamirror": "101010100100",
    "neapolitan pentachord 2": "101001100100",
    "schoenberg anagram hexachord": "111001100100",
    "pyramid": "101101100100",
    "debussy's heptatonic": "111101100100",
    "natural / genuine / 'black key' / blues pentatonic|slendro|bilahariraga": "101010010100",
    "indian dipaka|prometheus neapolitan": "110110010100",
    "indian|blues": "101110010100",
    "scriabin's mystic|prometheus hexachord": "110101010100",
    "dorian hexachord": "101101010100",
    "guidon / arezzo / natural / genuine / persian hexachord|quartal hexamirror": "101011010100",
    "messiaen's truncated 2|minor-bitonal hexachord": "110100110100",
    "dominant-11th|natural / genuine / lydian hexachord": "101010110100",
    "gypsy|moravian pistalkova (whistle)|alternating heptachord 1": "110110110100",
    "blues octatonic 1": "111110110100",
    "indian|chromatic inverse": "111001110100",
    "tritone major heptachord": "101011110100",
    "augmented|messiaen's truncated 3|lendvai's|genus tertium": "110011001100",
    "messiaen's truncated 2|petruskka chord|major-bitonal hexachord": "101100101100",
    "harmonic hexachord|augmented-11th|indian sviraga": "101010101100",
    "neapolitan": "111010101100",
    "harmonic minor|spanish gypsy": "110110101100",
    "persian|gypsy|hungarian|double harmonic|indian bhairava|turkish": "111001101100",
    "harmonic major": "110101101100",
    "diminished|alternating heptachord 2": "101101101100",
    "greek chromatic|indian": "101110011100",
    "modified blues": "101011011100",
    "messiaen's 4": "111100111100",
    "enigmatic heptatonic": "101010111100",
    "enigmatic octatonic|free-constructed": "111010111100",
    "blues octatonic 2": "101011111100",
    "chromatic decamirror": "111111111100",
    "whole-tone|messiaen's mode 1|raga gopriya|anhemitonic hexatonic": "101010101010",
    "neapolitan|leading whole-tone|combined|kafenda's": "111010101010",
    "jazz minor|bartok's|aug 13th heptamirror|acoustic|plane-altered": "110110101010",
    "greek|medieval|natural / genuine": "110101101010",
    "spanish octatonic": "110111101010",
    "greek complete|egyptian|blues|quartal octachord|diatonic octad": "111101011010",
    "spanish|major-minor|blues": "110111011010",
    "nonatonic blues": "111111011010",
    "messiaen's 6": "111010111010",
    "major-minor nonatonic|ramdasi malhar": "111101111010",
    "diminished|messiaen's 2|lendvai's": "110110110110",
    "diminished nonachord": "111011110110",
    "messiaen's 3|tsjerepnin": "111011101110",
    "major-minor mixed": "111111011110",
    "messiaen's 7": "111110111110",
    "chromatic undecamirror": "111111111110",
    "twelve-tone chromatic,dodecamirror": "111111111111",
}


def get_template_from_pitch_classes(pcs):
    """e.g. [0, 4, 7] -> [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0]"""
    template = [0] * 12

    for ix in pcs:
        template[ix] = 1

    return template


def get_template_from_midi_nums(midi_nums):
    """e.g. [48, 60] -> [1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]"""
    pitch_classes = [get_pitch_class_from_midi_num(n) for n in midi_nums]

    return get_template_from_pitch_classes(pitch_classes)


def get_template_from_template_str(template_str):
    """e.g. "101010101010" -> [1, 0, 1, 0, 1, 0, 1, 0, 1, 0, 1, 0]"""
    return [int(i) for i in list(template_str)]


def get_type_from_template(template):
    for chord_type in list(TYPE_TEMPLATES):
        if is_circular_match(
            template,
            get_template_from_template_str(TYPE_TEMPLATES[chord_type]),
        ):
            return chord_type
    return None


def get_type_from_type_num(num):
    """Takes the id of a chord type and returns its name"""
    return list(TYPE_TEMPLATES)[num]


def get_types_from_type_num_str(type_num_str):
    """Takes a type_num_str and returns a list of type names"""
    types = []

    for num in type_num_str.split("_"):
        if num == "":
            types.append("")
        else:
            types.append(get_type_from_type_num(int(num)))

    return types


def get_type_num_from_type(chord_type):
    """Takes the name of a chord type and returns its id"""
    if not chord_type:
        return None

    return list(TYPE_TEMPLATES).index(chord_type)


def notes_match_chord_type(notes, chord_type):
    """Returns true if any rotation of `notes` fit `chord_type`"""
    return is_circular_match(
        get_template_from_notes(notes),
        get_template_from_template_str(TYPE_TEMPLATES[chord_type]),
    )


def get_type_from_notes(notes):
    """
    Returns the first exact template match. Returns "" if no match.
    e.g. ["C4", "C3"] -> "unison"

    TODO: find partial matches as well
    """
    for chord_type in list(TYPE_TEMPLATES):
        if notes_match_chord_type(notes, chord_type):
            return chord_type

    logger.debug(f"No type template matched chord: {notes}")
    return ""


def get_type_from_midi_nums(midi_nums):
    """
    Returns the first exact template match
    e.g. [48, 60] -> "unison"
    """
    notes = [get_note_from_midi_num(n) for n in midi_nums]
    return get_type_from_notes(notes)


def get_template_from_notes(notes):
    """e.g. ["C4", "E4", "G4"] -> [1, 0, 0, 0, 1, 0, 0, 1, 0, 0, 0, 0]"""
    pitch_classes = [get_pitch_class_from_note(n) for n in notes]

    return get_template_from_pitch_classes(pitch_classes)



import warnings

import mido
import pretty_midi
from chord_progressions import DEFAULT_BPM, DEFAULT_MIDI_TICKS_PER_BEAT
from chord_progressions.midi import get_midi_ticks_from_seconds, make_midi_chord


def load_midi_file(filepath):
    # warnings can be verbose when midi has no metadata e.g. tempo, key, time signature
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)
        try:
            midi = pretty_midi.PrettyMIDI(filepath)
        except Exception as e:
            print(f"Failed loading file {filepath}: {e}")
            return

    return midi


def get_midi_from_chord(chord, name):
    mid = mido.MidiFile(type=0)
    mid.ticks_per_beat = DEFAULT_MIDI_TICKS_PER_BEAT

    track = mido.MidiTrack()
    track.name = name
    mid.tracks.append(track)

    # write a set tempo message with the bpm in microseconds per beat
    midi_tempo = mido.bpm2tempo(DEFAULT_BPM)
    track.append(mido.MetaMessage(type="set_tempo", tempo=midi_tempo))

    # write the chord progression
    tick_duration = get_midi_ticks_from_seconds(
        1, DEFAULT_BPM, DEFAULT_MIDI_TICKS_PER_BEAT
    )
    chord = make_midi_chord(chord, tick_duration)

    for msg in chord:
        track.append(msg)

    return mid


def get_midi_from_progression(progression, ticks_per_beat=DEFAULT_MIDI_TICKS_PER_BEAT):
    mid = mido.MidiFile(type=0)
    mid.ticks_per_beat = ticks_per_beat

    track = mido.MidiTrack()
    track.name = progression.name
    mid.tracks.append(track)

    # write a set tempo message with the bpm in microseconds per beat
    midi_tempo = mido.bpm2tempo(progression.bpm)
    track.append(mido.MetaMessage(type="set_tempo", tempo=midi_tempo))

    # write the chord progression
    for c, dur in zip(progression.chords, progression.get_durations_in_seconds()):
        tick_dur = get_midi_ticks_from_seconds(dur, progression.bpm, ticks_per_beat)
        chord = make_midi_chord(c, tick_dur)

        for msg in chord:
            track.append(msg)

    return mid



import json

from chord_progressions import DEFAULT_BPM, logger
from chord_progressions.chord import Chord
from chord_progressions.io.audio import make_audio_progression, save_audio_buffer
from chord_progressions.io.midi import get_midi_from_progression

"""
    Durations are specified in Tone.Time.Notation format.
    **NOT ALL TONE TIME NOTATIONS ARE SUPPORTED.**

    The number represents the subdivision. "t" represents a triplet. A "." adds a half.
    e.g. "4n" is a quarter note, "4t" is a quarter note triplet, and "4n." is a dotted quarter note.

    Tone.Time docs:
        https://github.com/Tonejs/Tone.js/wiki/Time

    Full list of Tone.js subdivisions:
        https://github.com/Tonejs/Tone.js/blob/dbed4d27fe66ee606a7309b03ac0ba4f5a2a4ecb/Tone/core/type/Units.ts#L47-L55s
"""
DURATIONS = ["8n", "4n", "4n.", "2n", "2n.", "1m", "2m"]

# The number of beats per duration
DURATION_BEAT_MAPPING = {
    "8n": 0.5,
    "4n": 1,
    "4n.": 1.5,
    "2n": 2,
    "2n.": 3,
    "1m": 4,
    "2m": 8,
}


def duration_to_seconds(duration, bpm):
    """Converts Tone.Time.notation to seconds"""
    beats = DURATION_BEAT_MAPPING[duration]
    seconds_per_beat = 60 / bpm
    return beats * seconds_per_beat


def get_closest(lst, K):
    """Returns the number from `lst` closest to a float `K`"""
    return min(range(len(lst)), key=lambda i: abs(lst[i] - K))


def seconds_to_duration(seconds, bpm):
    """Converts seconds to beats and quantizes to the closest supported beat notation"""
    beats_per_second = bpm / 60
    beats = seconds * beats_per_second
    ix = get_closest(list(DURATION_BEAT_MAPPING.values()), beats)
    return list(DURATION_BEAT_MAPPING.keys())[ix]


class Progression:
    """
    A sequence of Chords with metadata such as bpm.

    Parameters
    ----------
    chords: list[Chord] or list[list[int]], default []
        The chords in the progression.
    durations: list[str] or list[float], default []
        The duration of the chords, specified in Tone.Time notation or in seconds.
        Seconds will be quantized to Tone.Time units.
    bpm: float, default DEFAULT_BPM
        The beats per minute (BPM) of the progression.
    name: str, default ""
        The name of the progression.
    """

    def __init__(
        self,
        chords: list[Chord] = [],
        durations: list[str] = [],
        bpm: float = DEFAULT_BPM,
        name: str = "",
    ):
        if not isinstance(chords, list):
            raise ValueError("Chords must be a list")

        # If chords are passed as a midi_nums_list (list[list[int]]), convert them to Chords
        if isinstance(chords, list) and len(chords) > 0 and isinstance(chords[0], list):
            chords = [Chord(c) for c in chords]

        if not isinstance(durations, list):
            raise ValueError("Durations must be a list")

        if len(durations) == 0:
            durations = ["1m"] * len(chords)
        elif isinstance(durations[0], str):
            pass
        elif isinstance(durations[0], int) or isinstance(durations[0], float):
            durations = [seconds_to_duration(d, bpm) for d in durations]
        else:
            raise ValueError("Durations must be Tone.Time notation strings or seconds.")

        if not len(durations) == len(chords):
            raise ValueError("Durations must be the same length as chords.")

        self.chords = chords
        self.durations = durations
        self.bpm = bpm
        self.name = name
        self.metrics = {}  # TODO: add metrics

    def __iter__(self):
        # TODO: include progression-chord metadata
        for chord in self.chords:
            yield chord

    def __len__(self):
        return len(self.chords)

    def __repr__(self):
        return "Progression " + self.to_string()

    def __getitem__(self, ix):
        # TODO: include progression-chord metadata, e.g. its duration
        return self.chords[ix]

    def to_string(self):
        return json.dumps(self.to_json(), indent=4)

    def to_json(self):
        # TODO: include progression-level attrs like name, bpm
        result = []

        for ix, (chord, duration) in enumerate(list(zip(self.chords, self.durations))):
            result.append(
                {
                    "id": chord.id,
                    "ix": ix,
                    "duration": duration,
                    "type": chord.type,
                    "typeId": chord.typeId,
                    "notes": chord.notes,
                    "midi_nums": chord.midi_nums,
                    "metrics": chord.metrics,
                }
            )

        return result

    def get_durations_in_seconds(self):
        return [duration_to_seconds(d, self.bpm) for d in self.durations]

    def to_audio(self, outpath=None, n_overtones=4):
        dur_secs = self.get_durations_in_seconds()
        audio_buffer = make_audio_progression(self.chords, dur_secs, n_overtones)

        # TODO: create outpath from datetime if not provided? add flag to opt for this?
        if outpath:
            save_audio_buffer(audio_buffer, outpath)
            logger.info(f"Audio saved to {outpath}")

    def to_midi(self, outpath=None):
        mid = get_midi_from_progression(self)
        if outpath:
            mid.filename = outpath
            mid.save(outpath)
            logger.info(f"Midi saved to {outpath}")



from uuid import uuid4

from chord_progressions import logger
from chord_progressions.evaluate import evaluate_notes
from chord_progressions.io.audio import mk_chord_buffer, save_audio_buffer
from chord_progressions.io.midi import get_midi_from_chord
from chord_progressions.pitch import get_midi_num_from_note, get_note_from_midi_num
from chord_progressions.type_templates import (
    get_template_from_notes,
    get_type_from_midi_nums,
    get_type_num_from_type,
)

NoteList = list[str]
MidiNumList = list[int]


class Chord:
    """A set of unique notes.

    Parameters
    ----------
    notes: list[Chord], default []
        The set of notes in the chord. Can be specified as a list of midi numbers or as a list of note names.
            e.g. Chord([60, 64, 67])
            e.g. Chord(["C4", "E4", "G4"])
    """

    def __init__(self, notes: list = [], id: str = None):
        if isinstance(notes, list) and len(notes) > 0 and isinstance(notes[0], str):
            notes = [get_midi_num_from_note(n) for n in notes]

        self.init_from_midi_nums(midi_nums=notes, id=id)

    def init_from_midi_nums(self, midi_nums: MidiNumList = [], id: str = None):
        if any([i < 0 or i > 128 for i in midi_nums]):
            raise ValueError("The valid range of midi numbers is 0 to 128")

        midi_nums = sorted(list(set(midi_nums)))
        self.midi_nums = [int(i) for i in midi_nums]

        chord_type = get_type_from_midi_nums(midi_nums)
        self.type = chord_type
        self.typeId = get_type_num_from_type(chord_type)

        notes = [get_note_from_midi_num(n) for n in midi_nums]
        self.notes = notes
        self.metrics = evaluate_notes(notes)
        self.template = get_template_from_notes(notes)

        if not id:
            id = str(uuid4())
        self.id = id

    def __repr__(self):
        return "Chord " + self.to_string()

    def to_string(self):
        return str(self.to_json())

    def to_json(self):
        return {
            "id": self.id,
            "midi_nums": self.midi_nums,
            "type": self.type,
            "typeId": self.typeId,
            "notes": self.notes,
            "metrics": self.metrics,
        }

    def to_audio(self, outpath=None, n_overtones=4):
        audio = mk_chord_buffer(self.notes, 1, n_overtones)

        if outpath:
            save_audio_buffer(audio, outpath)
            logger.info(f"Audio saved to {outpath}")
        else:
            return audio

    def to_midi(self, outpath=None):
        mid = get_midi_from_chord(self.chords, 1, 120, "-".join(self.notes))

        if outpath:
            mid.filename = outpath
            mid.save(outpath)
            logger.info(f"Midi saved to {outpath}")
        else:
            return mid



from chord_progressions.utils import is_circular_match, shift_arr_by_one


def test_shift_arr_by_one():

    assert shift_arr_by_one([1, 2, 3]) == [3, 1, 2]

    template = [1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0]
    expected = [0, 1, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0]
    assert shift_arr_by_one(template) == expected

    assert shift_arr_by_one([1]) == [1]


def test_is_circular_match():
    assert is_circular_match([1, 2, 3], [3, 1, 2])
    assert is_circular_match([0, 0, 0], [0, 0, 0])
    assert is_circular_match([0], [0])



import logging
import os

__version__ = "0.37.0"  # updated by bumpversion, do not change

THIS_DIR = os.path.dirname(os.path.realpath(__file__))

OUTPUT_DIR = os.path.join(THIS_DIR, "../assets/output")

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger("chord-progressions")

# Config
# TODO: move to config
DEFAULT_BPM = 120
DEFAULT_MIDI_TICKS_PER_BEAT = 480

from chord_progressions.chord import Chord
from chord_progressions.extract import (
    extract_progression_from_midi,
)
from chord_progressions.pitch import (
    get_midi_num_from_note,
    get_note_from_midi_num,
    get_note_list,
    get_pitch_class_from_note,
    get_midi_nums_list_from_midi_nums_str,
    get_notes_list_from_midi_nums_str,
)
from chord_progressions.progression import Progression, DURATIONS
from chord_progressions.type_templates import (
    TYPE_TEMPLATES,
    get_template_from_template_str,
    get_template_from_notes,
    get_type_from_notes,
    get_type_num_from_type,
)

# Use __all__ to let type checkers know what is part of the public API
__all__ = [
    Chord,
    extract_progression_from_midi,
    get_midi_num_from_note,
    get_note_from_midi_num,
    get_note_list,
    get_pitch_class_from_note,
    get_midi_nums_list_from_midi_nums_str,
    get_notes_list_from_midi_nums_str,
    Progression,
    DURATIONS,
    TYPE_TEMPLATES,
    get_template_from_template_str,
    get_template_from_notes,
    get_type_from_notes,
    get_type_num_from_type,
]



import pytest
from chord_progressions.chord import Chord


def test_instantiate_chord():
    assert Chord().midi_nums == []

    c = Chord(["C4", "E4", "G4"])
    assert c.midi_nums == [60, 64, 67]

    with pytest.raises(ValueError):
        assert Chord([-1])

    cd = Chord(["C4", "E4", "G4"], 2)
    assert cd.midi_nums == [60, 64, 67]



from chord_progressions.evaluate import get_evenness, get_interval_class_vector

EXPECTATIONS = [
    {"notes": ["C4"], "icv": [0, 0, 0, 0, 0, 0], "evenness": 0},  # unison
    {"notes": ["C4", "C#4"], "icv": [1, 0, 0, 0, 0, 0], "evenness": 0.51},  # semitone
    {
        "notes": ["C4", "D#4"],  # minor third
        "icv": [0, 0, 1, 0, 0, 0],
        "evenness": 1.41,
    },
    {
        "notes": ["C4", "E4"],  # major third
        "icv": [0, 0, 0, 1, 0, 0],
        "evenness": 1.73,
    },
    {
        "notes": ["D4", "A5"],  # perfect fourth
        "icv": [0, 0, 0, 0, 1, 0],
        "evenness": 1.93,
    },
    {
        "notes": ["D4", "E5", "D#6", "F#7"],  # major-second tetracluster 2
        "icv": [2, 2, 1, 1, 0, 0],
        "evenness": 6.18,
    },
    {
        "notes": ["F2", "B2", "D3", "G#3"],  # diminished seventh
        "icv": [0, 0, 4, 0, 0, 2],
        "evenness": 9.66,
    },
]


def test_get_interval_class_vector():
    for ex in EXPECTATIONS:
        assert get_interval_class_vector(ex["notes"]) == ex["icv"]


def test_get_evenness():
    tolerance = 0.01

    for ex in EXPECTATIONS:
        assert get_evenness(ex["icv"]) - ex["evenness"] < tolerance



"""
NOTE: experimental. Essentia not included in dependencies.

Implementation details:

For each frame:
    1. Obtain spectrogram with STFT
    2. Detect spectral peaks
    3. Estimate HPCP [Gómez 2006]

Join frames together into segments corresponding to each beat of the track.

For each segment:
    4. Avg and normalize HPCPs
    5. Match avg'd HPCPs to Chordico's tone profiles
"""

import os

import essentia
import essentia.standard as es
import essentia.streaming as ess
import matplotlib.pyplot as plt
import numpy as np
from chord_progressions import logger
from chord_progressions.chord import Chord
from chord_progressions.progression import Progression
from pylab import imshow

SAMPLE_RATE = 44100  # Hz
FRAME_SIZE = 4096  # samples
HPCP_THRESH = 0.35

PC_MIDI_NUMS = list(range(69, 81))


def get_essentia_features(filepath):
    # TODO: preprocess audio to eliminate regions where harmonic structure is noisy
    #   e.g. source-separate guitar from the rest of the mix?

    # Initialize the essentia streaming algorithms
    loader = ess.MonoLoader(filename=filepath)

    framecutter = ess.FrameCutter(
        frameSize=FRAME_SIZE, hopSize=int(FRAME_SIZE / 2), silentFrames="noise"
    )

    windowing = ess.Windowing(type="blackmanharris62")

    spectrum = ess.Spectrum()

    spectralpeaks = ess.SpectralPeaks(
        orderBy="magnitude",
        magnitudeThreshold=0.00001,
        minFrequency=20,
        maxFrequency=3500,
        maxPeaks=60,
    )

    hpcp_extractor = ess.HPCP(
        size=12,  # the size of the output HPCP (must be a positive nonzero multiple of 12)
        referenceFrequency=440,  # the reference frequency for semitone index calculation, corresponding to A3 [Hz]
        minFrequency=20,  # the min freq that contributes to the HPCP [Hz]
        maxFrequency=3500,  # the max freq that contributes to the HPCP [Hz]
        weightType="cosine",  # type of weighting function for determining frequency contribution
        nonLinear=False,  # apply non-linear mapping to the output (boost values close to 1, decrease values close to 0)
        windowSize=1.0,  # the size, in semitones, of the window used for the weighting
        bandPreset=False,  # don't use a band preset
    )

    # Use a pool to store data
    pool = essentia.Pool()

    # Connect streaming algorithms
    loader.audio >> framecutter.signal
    loader.audio >> (pool, "audio")

    framecutter.frame >> windowing.frame >> spectrum.frame
    spectrum.spectrum >> spectralpeaks.spectrum
    spectralpeaks.magnitudes >> hpcp_extractor.magnitudes
    spectralpeaks.frequencies >> hpcp_extractor.frequencies

    # TODO: use a representation w voicings rather than just HPCPs (use pitches, not just pitch classes)
    hpcp_extractor.hpcp >> (pool, "hpcp")

    # TODO: compute [HighResolutionFeatures](https://essentia.upf.edu/reference/std_HighResolutionFeatures.html)
    # TODO: compute hpcp crest & entropy (https://essentia.upf.edu/tutorial_extractors_musicextractor.html)
    # TODO: get all sorts of other features w MusicExtractor (https://essentia.upf.edu/streaming_extractor_music.html)

    # Run the streaming network
    essentia.run(loader)

    feats = {i: pool[i] for i in pool.descriptorNames() if i != "audio"}

    return pool["audio"], feats


def plot_hpcp(hpcp, figpath):
    plt.rcParams["figure.figsize"] = (15, 6)
    imshow(hpcp.T, aspect="auto", origin="lower", interpolation="none")
    plt.yticks(
        ticks=range(12),
        labels=["A", "A#", "B", "C", "C#", "D", "D#", "E", "F", "F#", "G", "G#"],
    )
    plt.title("HPCPs in frames")
    plt.savefig(figpath)
    plt.clf()
    print(f"Saved HPCP plot to {figpath}")


def get_beats(x, marked_audio_path=None):
    # Compute beat positions and BPM
    rhythm_extractor = es.RhythmExtractor2013(method="multifeature")
    bpm, beats, beats_confidence, _, beats_intervals = rhythm_extractor(x)
    print(f"  BPM: {bpm}")

    if marked_audio_path:
        # Mark beat positions in the audio and write it to a file
        marker = es.AudioOnsetsMarker(onsets=beats, type="noise")
        marked_audio = marker(x)
        es.MonoWriter(filename=marked_audio_path)(marked_audio)

    return beats, bpm


def get_meter(signal, beats):
    """Estimate the time signature of a signal by finding the highest correlation between salient beat loudness bands"""
    loudness, loudness_band_ratio = es.BeatsLoudness(
        beats=beats,  # list of beat positions (in seconds)
        beatWindowDuration=0.1,  # seconds, dur of window in which to look for beat start (centered on beat positions)
        beatDuration=0.05,  # duration of window in which the beat will be restricted [s]
        frequencyBands=[
            20,
            150,
            400,
            3200,
            7000,
            22000,
        ],  # the list of bands to compute energy ratios, hz
        sampleRate=SAMPLE_RATE,
    )(signal)

    beatogram = es.Beatogram(size=16)(  # size: number of beats for dynamic filtering
        loudness, loudness_band_ratio
    )

    return es.Meter()(beatogram)


def extract_progression_from_audio(filepath, plot_dir=None, marked_audio_dir=None):
    """
    Params
        filepath: str, Path to an audio file
        plot_dir: str, if passed, writes intermediate plots. Useful for debugging.
        marked_audio_dir: str, writes an audio file with beat position markers. Useful for debugging.

    Returns
        progression: Progression, the extracted progression object
    """

    basename = os.path.basename(filepath)
    trackid = os.path.splitext(basename)[0]

    x, feats = get_essentia_features(filepath)

    hpcp = feats["hpcp"]
    logger.info(f"  HPCP shape: {hpcp.shape}")

    hpcp_figpath = os.path.join(plot_dir, f"{trackid}_hpcp.png")
    plot_hpcp(hpcp, hpcp_figpath)

    marked_audio_path = os.path.join(marked_audio_dir, f"{trackid}_beats.mp3")
    beats, bpm = get_beats(x, marked_audio_path)

    meter = get_meter(x, beats)
    print(f"  Estimated meter: {meter}")

    segment_start_times = beats[:: int(meter)]
    frames_per_second = SAMPLE_RATE / FRAME_SIZE
    quarter_note_beat_samples = [
        int(i * frames_per_second) for i in segment_start_times
    ]
    beat_samples = list(
        zip(quarter_note_beat_samples[:-1], quarter_note_beat_samples[1:])
    )
    beat_durations_samples = [b[1] - b[0] for b in beat_samples]
    beat_durations = [i / frames_per_second for i in beat_durations_samples]

    segment_pcs = []

    for fstart, fend in beat_samples:
        frame = hpcp[fstart:fend]
        mean = frame.mean(axis=0)
        normalized_mean = mean / mean.max() if mean.max() > 0 else mean
        pcs = np.where(normalized_mean > HPCP_THRESH)[0].tolist()
        segment_pcs.append(pcs)

    chords = [Chord([PC_MIDI_NUMS[pc] for pc in seg]) for seg in segment_pcs]
    # TODO: convert beat_durations to Tone Time values relative to bpm and add them to Progression
    progression = Progression(chords, beat_durations, bpm=bpm, name=filepath)

    return progression



"""
    Defines the metrics computed for every progression.

    TODO:
        - A common heuristic is that an interval is consonant when the ratio between the frequencies is simple.
        - Resonance
        - Brilliance: "for maximum brilliance let the lower tones [be somehow equally spaced out as] their overtones"
            - Vincent Persechetti, Twentieth-Century Harmony
        - something like https://en.xen.wiki/w/Harmonic_Entropy
"""

import itertools

import numpy as np
from chord_progressions.pitch import (
    get_freq_from_note,
    get_midi_num_from_note,
    get_n_overtones_harmonic,
    get_pitch_class_from_note,
)
from chord_progressions.type_templates import get_template_from_notes


def get_interval_class_vector(notes):
    """
    Returns a vector of interval classes.
        0: minor second / major seventh (1 or 11 semitones)
        1: major second / minor seventh (2 or 10 semitones)
        2: minor third / major sixth (3 or 9 semitones)
        3: major third / minor sixth (4 or 8 semitones)
        4: perfect fourth / perfect fifth (5 or 7 semitones)
        5: tritone (6 semitones)
    """
    vec = [0] * 6

    template = get_template_from_notes(notes)
    one_indices = [ix for ix, i in enumerate(template) if i == 1]
    pairs = list(itertools.combinations(one_indices, 2))

    intervals = [p[1] - p[0] for p in pairs]

    modded = [12 - i if i > 6 else i for i in intervals]

    for i in range(6):
        vec[i] = modded.count(i + 1)

    return vec


def get_evenness(interval_class_vector):
    """
    A rough measure of acoustic consonance. Highly consonant chords divide the octave nearly evenly.
    Note that acoustic consonance implies near-evenness, but not the reverse.
    For background on the equation see https://www.researchgate.net/profile/Jack_Douthett/publication/249881698_Vector_Products_and_Intervallic_Weighting/links/575061d708ae1c34b39aaa1b.pdf
    """
    weight_vector = [0] * 6

    for i in range(6):

        k = i + 1

        weight_vector[i] += 2 * np.sin(k * np.pi / 12)

    weight = np.dot(weight_vector, interval_class_vector)

    return weight


def get_relative_evenness(evenness, cardinality):
    """
    Calculate the evenness of a pitch class set relative to its possible range.
    A relative value is useful because highly-even dyads will have a lower evenness than highly uneven hexachords.
    The result is a value between 0 and 1.
    """
    # TODO: replace this table with a formula
    MIN_MAX_EVENNESS_BY_CARDINALITY = [
        [0.0, 0.0],
        [0.5176380902050415, 2.0],
        [2.035276180410083, 5.196152422706632],
        [6.1815405503520555, 9.65685424949238],
        [9.631030293135233, 15.287884542627612],
        [16.226784405860382, 22.392304845413264],
        [24.822538518585535, 30.479392768077915],
        [35.350144283888824, 40.03987070039298],
        [47.609800856760984, 50.77067709905754],
        [61.28367099200624, 62.7660329018012],
        [75.9575411272515, 75.9575411272515],
        [91.14904935270181, 91.14904935270181],
    ]

    emin, emax = MIN_MAX_EVENNESS_BY_CARDINALITY[cardinality - 1]
    return (evenness - emin) / (emax - emin)


def get_overtone_agreement(notes):

    freqs = [get_freq_from_note(n) for n in notes]

    all_partials = []

    n_overtones = 10
    for f in freqs:
        all_partials.extend(get_n_overtones_harmonic(f, n_overtones))

    sorted_partials = sorted(all_partials)

    sum_of_differences = 0

    for i in range(1, len(sorted_partials)):
        diff = sorted_partials[i] - sorted_partials[i - 1]
        sum_of_differences += diff

    return sum_of_differences


def evaluate_notes(notes):
    """list of notes, e.g. ["C4", "E4", "G4"]"""

    metrics = {}

    try:

        interval_class_vector = get_interval_class_vector(notes)

        evenness = get_evenness(interval_class_vector)

        pc_cardinality = len(set([get_pitch_class_from_note(n) for n in notes]))
        assert pc_cardinality <= 12, "Pitch class cardinality > 12"

        metrics["num_notes"] = len(notes)
        metrics["num_pitches"] = len(set(notes))
        metrics["pc_cardinality"] = pc_cardinality
        metrics["interval_class_vector"] = interval_class_vector
        metrics["ambitus"] = get_ambitus(notes)
        metrics["evenness"] = evenness
        # metrics["relative_evenness"] = get_relative_evenness(evenness, pc_cardinality)
        # metrics["overtone_agreement"] = get_overtone_agreement(notes)

    except:
        pass

    return metrics


def get_macroharmony(progression):
    """The total collection of notes"""

    collection = set()

    for notes in progression:
        for note in notes:
            collection.add(note)

    return collection


def get_ambitus(macroharmony):
    """Ambitus <int>: the number of semitones between the lowest and highest note"""
    ambitus = 0

    midi_notes = [get_midi_num_from_note(n) for n in macroharmony]
    if midi_notes:
        ambitus = max(midi_notes) - min(midi_notes)

    return ambitus


def evaluate_progression(progression):
    metrics = {}

    # try:
    #     macroharmony = get_macroharmony(notes_list)
    #     metrics["ambitus"] = get_ambitus(macroharmony)
    #     metrics["density"] = len(macroharmony) / len(notes_list)
    # except:
    #     pass

    return metrics


def get_n_overtones_for_notes(notes, n):
    """Calculates `n` overtone frequencies for each note in `notes`"""

    freqs = [get_freq_from_note(n) for n in notes]
    o_freqs = [get_n_overtones_harmonic(f, n) for f in freqs]

    return o_freqs



import mido
from chord_progressions import DEFAULT_MIDI_TICKS_PER_BEAT


def make_midi_chord(chord, end_tick: int):
    note_onsets = []
    note_offsets = []

    first_note = True
    for midi_num in chord.midi_nums:
        note_onsets.append(mido.Message("note_on", note=midi_num, time=0))

        if not first_note:
            end_tick = 0

        note_offsets.append(mido.Message("note_off", note=midi_num, time=end_tick))
        first_note = False

    return note_onsets + note_offsets


def get_midi_ticks_from_seconds(
    seconds, bpm, ticks_per_beat=DEFAULT_MIDI_TICKS_PER_BEAT
):
    midi_tempo = mido.bpm2tempo(bpm)
    return int(mido.second2tick(seconds, ticks_per_beat, midi_tempo))


def get_seconds_from_midi_ticks(ticks, bpm, ticks_per_beat=DEFAULT_MIDI_TICKS_PER_BEAT):
    midi_tempo = mido.bpm2tempo(bpm)
    return mido.tick2second(ticks, ticks_per_beat, midi_tempo)



import numpy as np
from chord_progressions import logger
from chord_progressions.pitch import get_freq_from_note, get_n_overtones_harmonic
from scipy.io import wavfile

SAMPLE_RATE = 44100


def mk_sin(freq, duration, amp):
    s = [
        amp * np.sin(n * 2 * np.pi * freq / SAMPLE_RATE)
        for n in range(int(duration * SAMPLE_RATE))
    ]
    return np.array(s).reshape(-1, 1)


def mk_freq_buffer(freq, duration, n_overtones):
    freqs = get_n_overtones_harmonic(freq, n_overtones)

    # make each overtone progressively quieter
    sins = [mk_sin(f, duration, 1 / 2**i) for i, f in enumerate(freqs)]

    # TODO: normalize such that all notes have the same amplitude
    # regardless of the number of frequencies
    normed_sins = [s / len(sins) for s in sins]

    # combine the overtones to make a note
    return sum(normed_sins)


def mk_note_buffer(note, duration, n_overtones):

    freq = get_freq_from_note(note)

    return mk_freq_buffer(freq, duration, n_overtones)


def combine_buffers(buffers):
    stacked = np.hstack(buffers)

    normed = stacked / len(buffers)

    return np.sum(normed, axis=1).astype("float32")


def mk_chord_buffer(chord, duration, n_overtones):
    logger.debug(f"Generating {duration} second buffer for chord: {chord}")

    note_buffers = [mk_note_buffer(note, duration, n_overtones) for note in chord]

    return combine_buffers(note_buffers)


def mk_arpeggiated_chord_buffer(chord, duration, seqs, n_overtones):

    logger.debug(f"Generating {duration} second arpeggiated buffer for chord: {chord}")

    selected_seqs = [seqs[np.random.randint(len(seqs))] for i in range(len(chord))]

    bufs = []

    for note, seq in zip(chord, selected_seqs):
        logger.debug(f"  {note} got rhythm {''.join(map(str, seq))}")

        pos_dur = duration / len(seq)

        seq_bufs = []

        for pos in seq:

            if pos == 1:
                seq_bufs.append(mk_note_buffer(note, pos_dur, n_overtones))

            else:
                seq_bufs.append(mk_sin(0, pos_dur, 0))

        bufs.append(np.concatenate(seq_bufs))

    # TODO: would these 0s be better placed somewhere in the middle of the bufs?
    max_buf_len = max([b.shape[0] for b in bufs])

    # TODO: can be simplified
    for ix, buf in enumerate(bufs):
        padding = np.zeros((max_buf_len - len(buf), 1))
        bufs[ix] = np.concatenate([buf, padding])

    return np.sum(bufs, axis=0)


def save_sample_chord_audio(chord, chord_type):
    buf = mk_chord_buffer(chord, duration=1, n_overtones=1)
    filename = chord_type.replace("/", "_").replace(" ", "")
    outpath = f"sample_chords/{filename}.wav"
    save_audio_buffer(buf, outpath)


def make_audio_progression(chords, durations, n_overtones):
    buffers = [
        mk_chord_buffer(c.notes, d, n_overtones) for c, d in zip(chords, durations)
    ]

    return np.concatenate(buffers)


def save_audio_buffer(buf, outpath):
    wavfile.write(outpath, SAMPLE_RATE, buf)



import pytest
from chord_progressions.chord import Chord
from chord_progressions.progression import (
    Progression,
    duration_to_seconds,
    seconds_to_duration,
)

TWO_CHORDS = [Chord(["C2", "A4", "G6"]), Chord(["C-1", "A4", "G9"])]


def test_instantiate_progression():
    p = Progression(chords=TWO_CHORDS)
    assert len(p) == len(TWO_CHORDS)

    failing_inputs = [
        "a",
        [1, 2, 3],
        [[1, 2, 3], [4, 5, 6]],
        ["E4", "C4"],
        [["E4", "C4"], ["G10"]],
        [["E4", "C4"], "G4"],
    ]
    with pytest.raises(Exception):
        for arg in failing_inputs:
            assert Progression(arg)

    valid_inputs = [[["E4", "C4"], ["G4", "A4", "B5"]]]
    for arg in valid_inputs:
        assert Progression(arg)

    p = Progression(["E4", "C4"], durations=["1m", "2m"])
    assert p.durations == ["1m", "2m"]


def test_duration_to_seconds():
    assert duration_to_seconds("4n", 120) == 0.5
    assert duration_to_seconds("2n", 120) == 1.0
    assert duration_to_seconds("1m", 120) == 2.0


def test_seconds_to_duration():
    assert seconds_to_duration(0.0001, 120) == "8n"
    assert seconds_to_duration(0.5, 120) == "4n"
    assert seconds_to_duration(1, 120) == "2n"
    assert seconds_to_duration(1.5, 120) == "2n."
    assert seconds_to_duration(2.0, 120) == "1m"
    assert seconds_to_duration(4.0, 120) == "2m"
    assert seconds_to_duration(100.0, 120) == "2m"



"""
Implements a method for extracting chords from symbolic music that is based on
the HarmAn algorithm by Pardo & Birmingham:

https://interactiveaudiolab.github.io/assets/papers/pardo-birmingham-cmj02.pdf
"""

import os
import time
from collections import Counter, defaultdict
from copy import deepcopy

import networkx as nx
import numpy as np
import pandas as pd
from chord_progressions import logger
from chord_progressions.chord import Chord
from chord_progressions.extract.midi import (
    DEFAULT_QUANTIZE_BEAT,
    DEFAULT_SHORTEST_NOTE,
    DEFAULT_SMOOTH_BEAT,
)
from chord_progressions.extract.midi import simplify_harmony
from chord_progressions.io.midi import load_midi_file
from chord_progressions.pitch import (
    get_note_from_midi_num,
    get_pitch_class_from_midi_num,
)
from chord_progressions.progression import Progression
from chord_progressions.type_templates import (
    TYPE_TEMPLATES,
    get_template_from_template_str,
    get_type_num_from_type,
)
from chord_progressions.utils import shift_arr_by_one

# We pre-calculate all the templates we want to use as labels
# TEMPLATE_LABELS will be {chord_type_id: (rotations, one_indices)}
MIN_NUM_NOTES = 3
MAX_NUM_NOTES = 6
TEMPLATE_LABELS = {}


def get_template_str(template):
    return "".join([str(i) for i in template])


def get_all_rotations_of_template(template):
    rotation_strs = set()
    rotation_strs.add(get_template_str(template))

    prev = template

    for i in range(len(template) - 1):
        rotated = shift_arr_by_one(prev)
        rotation_strs.add(get_template_str(rotated))

        prev = rotated

    return [[int(j) for j in list(i)] for i in rotation_strs]


for template_name, template_str in TYPE_TEMPLATES.items():

    template_id = get_type_num_from_type(template_name)
    template = get_template_from_template_str(template_str)

    if sum(template) >= MIN_NUM_NOTES and sum(template) <= MAX_NUM_NOTES:
        if template_id not in TEMPLATE_LABELS:
            TEMPLATE_LABELS[template_id] = []

        rotations = get_all_rotations_of_template(template)

        one_indices = []
        for rotation in rotations:
            one_indices.append([ix for ix, i in enumerate(rotation) if i == 1])

        TEMPLATE_LABELS[template_id] = (rotations, one_indices)


class PartitionPoint:
    """A partition point occurs where the set of sounding notes changes by the onset or offset of one or more notes."""

    def __init__(self, start):
        self.time = start
        self.onsets = []
        self.offsets = []

    def add_event(self, note, is_onset):
        if is_onset:
            self.onsets.append(note)
        else:
            self.offsets.append(note)

    def pprint(self):
        logger.debug(f"    time {self.time}")
        logger.debug(f"    onsets {[get_note_from_midi_num(i) for i in self.onsets]}")
        logger.debug(
            f"    offsets, {[get_note_from_midi_num(i) for i in self.offsets]}"
        )


class MinimalSegment:
    """A minimal segment is the interval between two sequential partition points."""

    def __init__(self):
        self.start = 0  # seconds
        self.end = 0  # seconds
        self.midi_nums = []  # midi note numbers
        self.template_scores = {}

    def update_midi_nums(self, midi_nums, add):
        if add:
            self.midi_nums.extend(midi_nums)
        else:
            for p in midi_nums:
                self.midi_nums.remove(p)

    def get_pitch_classes(self):
        return [get_pitch_class_from_midi_num(p) for p in self.midi_nums]

    def get_pitch_class_weights(self):
        pcs = self.get_pitch_classes()
        return {pc: pcs.count(pc) for pc in pcs}

    def get_template_scores(self):
        pcs = self.get_pitch_classes()
        pc_weights = {pc: pcs.count(pc) for pc in pcs}
        return get_template_scores(pc_weights)

    def get_notes(self):
        return [get_note_from_midi_num(n) for n in self.midi_nums]

    def get_note_occurences(self):
        midi_nums = self.midi_nums
        return {n: midi_nums.count(n) for n in midi_nums}

    def pprint(self):
        logger.debug(f"    start {self.start}")
        logger.debug(f"    end {self.end}")
        logger.debug(f"    midi_nums {self.get_notes()}")


def parse_events(pmid):
    """Parses a PrettyMIDI object into a 2D array with (time, note, is_onset) columns."""

    points = []

    for instrument in pmid.instruments:

        if instrument.is_drum:
            continue

        for note in instrument.notes:
            points.append((note.start, note.pitch, 1))
            points.append((note.end, note.pitch, 0))

    dtype = np.dtype([("time", float), ("note", int), ("is_onset", bool)])
    events = np.array(points, dtype=dtype)

    return np.sort(events, order="time")


def get_partition_points(events):
    """Returns the partition points for a set of events, ordered by time."""

    p_all = []

    time_cursor = events[0][0]
    point = PartitionPoint(time_cursor)

    for event in events:
        t = event["time"]
        note = event["note"]
        is_onset = event["is_onset"]

        if time_cursor == t:
            point.add_event(note, is_onset)

        else:
            p_all.append(deepcopy(point))

            time_cursor = t

            point = PartitionPoint(t)
            point.add_event(note, is_onset)

    p_all.append(deepcopy(point))

    # Uncomment for debug output
    logger.debug(f"--- {len(p_all)} partition points ---")
    for ix, p in enumerate(p_all):
        logger.debug(f"  p{ix}")
        p.pprint()

    return p_all


def get_segment_pc_weights(segment):
    """Aggregates the pitch class weights of all minimal segments within a segment."""

    segment_pc_weights = defaultdict(int)

    for s in segment:
        pc_weights = s.get_pitch_class_weights()

        for pc, weight in pc_weights.items():
            print(pc, weight)
            segment_pc_weights[pc] += weight

    return segment_pc_weights


def get_template_score(pc_weights, template, oix):
    """Scores a weighted set of pitch classes against a single template.

    Uses the algorithm defined in Pardo 2002 Fig 4.

    Input:
        pc_weights, a dict of {pc: weight}
        template, the template against which to score. Example: [1, 0, 0, 0, 1, 0, 0, 0, 0, 1, 0, 0]
        oix, the one-indices for the template, pre-computed for performance. Example: [0, 4, 9]

    Returns:
        S, an integer score
    """

    P = 0  # positive evidence
    N = 0  # negative evidence
    M = 0  # misses

    for pc, weight in pc_weights.items():
        if template[pc]:
            P += weight
        else:
            N += weight

    for _ in set(oix) - set(pc_weights):
        M += 1

    return P - (M + N)


def get_template_scores(pc_weights):
    """Scores a weighted set of pitch classes against all template labels.

    Returns a dict of {label: score}
    """
    scores = {}

    for template_name, (rotations, one_indices) in TEMPLATE_LABELS.items():

        rotation_scores = []

        for ix, rotation in enumerate(rotations):
            score = get_template_score(pc_weights, rotation, one_indices[ix])

            rotation_scores.append(score)

        # TODO: use dim7 resolution as tiebreaker (see Fig 4 in Pardo 2002)
        best_rotation_score = max(rotation_scores)

        if template_name not in scores:
            scores[template_name] = best_rotation_score
        else:
            scores[template_name] += best_rotation_score

    return scores


def get_minimal_segments(p_all):
    """Returns the minimal segments for a set of partition points, ordered by time."""

    s_m = []

    s = MinimalSegment()

    s.start = p_all[0].time
    s.update_midi_nums(p_all[0].onsets, 1)

    for point in p_all[1:]:
        s.end = point.time

        midi_nums = s.midi_nums

        s_m.append(deepcopy(s))

        s = MinimalSegment()

        s.start = point.time

        s.update_midi_nums(midi_nums, 1)
        s.update_midi_nums(point.onsets, 1)

        s.update_midi_nums(point.offsets, 0)

    logger.debug(f"\n--- {len(s_m)} minimal segments ---")

    ## Uncomment for debug output
    # for ix, sm in enumerate(s_m):
    #     logger.debug(f"  sm{ix}")
    #     # sm.pprint()
    #     pc_weights = sm.get_pitch_class_weights()
    #     sm.template_scores = get_template_scores(pc_weights)

    return s_m


def get_segment_label(segment):
    """Aggregates the template scores of all minimal segments within a segment and returns the best label.

    See Figure 3 in Pardo 2002
    """

    if not len(segment) > 0:
        raise ValueError("Empty segment")

    counter = Counter(segment[0].get_template_scores())

    for ms in segment[1:]:
        counter.update(Counter(ms.get_template_scores()))

    if not len(counter) > 0:
        raise ValueError("No scores for segment")

    # TODO: it's common for multiple templates to have the same score. How should we tiebreak?
    best_template, best_score = counter.most_common(1)[0]

    return best_score, best_template


def get_segment_midi_nums(segment):
    """Aggregates the notes in all the minimal segments that comprise a segment"""

    midi_nums = []

    for ms in segment:
        midi_nums.extend(ms.get_note_occurences())

    return midi_nums


def segment_and_label(p_all, s_m):
    """
    Based on the HarmAn algorithm by Pardo & Birmingham:
        https://interactiveaudiolab.github.io/assets/papers/pardo-birmingham-cmj02.pdf

    Implementations to reference:
        - https://github.com/salilgupta1/chordal_analysis/blob/c290dda5684c1d63d37867abdf0008cff41b2e5f/chordal_analysis/main.py
        - https://github.com/tomthecollins/wimir-eurovision/blob/68b1fa8bc616b751c0681dcc07fe0dc3079caabb/public/lib/maia-util_r0.2.12.js#L2817
    """

    logger.debug("\n--- Segmentation ---")

    t0 = time.time()

    # in the graph, nodes are partition points and edges are segments
    G = nx.DiGraph()

    final_ix = len(p_all) - 1

    ui = 0  # index of the preceding vertex

    for vi, p in enumerate(p_all):

        if vi in [0, final_ix]:
            G.add_node(vi, marked=1, time=p.time)
            continue

        uv_segment = s_m[ui:vi]  # preceding segment
        uv_midi_nums = get_segment_midi_nums(uv_segment)
        uv_notes = [get_note_from_midi_num(n) for n in uv_midi_nums]

        # a minimal segment without notes is a rest for all instruments
        if len(uv_midi_nums) == 0:
            logger.debug(f"  {vi}/{final_ix} - Continue: No notes in segment")
            continue

        uv_score, uv_label = get_segment_label(uv_segment)

        wi = vi + 1  # index of the succeeding vertex
        vw_score, vw_label = get_segment_label(s_m[vi:wi])  # succeeding segment
        uw_score, uw_label = get_segment_label(s_m[ui:wi])  # full segment

        if uw_score < (uv_score + vw_score):
            logger.debug(
                f"  {vi}/{final_ix} - New: {uw_label} ({uw_score}) < {uv_label} ({uv_score}) + {vw_label} ({vw_score})"
            )

            edge_attributes = {
                "start_segment": ui,
                "end_segment": vi,
                "label": uv_label,
                "score": uv_score,
                "midi_nums": uv_midi_nums,
                "notes": uv_notes,
            }

            G.add_node(vi, marked=1, time=p.time)
            G.add_edge(ui, vi, **edge_attributes)

            ui = vi
        else:
            logger.debug(
                f"  {vi}/{final_ix} - Continue: {uw_label} ({uw_score}) >= {uv_label} ({uv_score}) + {vw_label} ({vw_score})"
            )

    final_segment = s_m[ui:final_ix]
    final_segment_midi_nums = get_segment_midi_nums(uv_segment)
    final_segment_notes = [get_note_from_midi_num(n) for n in final_segment_midi_nums]

    final_score, final_label = get_segment_label(final_segment)

    final_edge_attributes = {
        "start_segment": ui,
        "end_segment": final_ix,
        "label": final_label,
        "score": final_score,
        "midi_nums": final_segment_midi_nums,
        "notes": final_segment_notes,
    }

    G.add_edge(ui, final_ix, **final_edge_attributes)

    logger.debug("\n--- Results ---")
    logger.debug(f"{len(G.edges)} segments")

    rows = []

    template_names = list(TYPE_TEMPLATES)

    for e in G.edges:
        edge = G[e[0]][e[1]]

        start = G.nodes[e[0]]["time"]
        end = G.nodes[e[1]]["time"]

        row = {
            "start_segment": edge["start_segment"],
            "start_time": start,
            "end_segment": edge["end_segment"],
            "end_time": end,
            "duration": end - start,
            "label_id": edge["label"],
            "label": template_names[edge["label"]],
            "midi_nums": edge["midi_nums"],
            "notes": edge["notes"],
            "chord": Chord(edge["midi_nums"]),
        }
        rows.append(row)

    logger.debug(f"segment_and_label done in {time.time() - t0} seconds")

    return rows


def label_midi(midi):
    events = parse_events(midi)

    p_all = get_partition_points(events)

    s_m = get_minimal_segments(p_all)

    return segment_and_label(p_all, s_m)


def label_file(filepath):
    logger.debug(f"Labeling {filepath}")

    midi = load_midi_file(filepath)

    if not midi:
        return None

    return label_midi(midi)


def write_labels(labels, inpath, outpath, arrangement_id=None):
    if not labels:
        logger.error(f"Failed for {inpath}")
        return {inpath: None}

    if not arrangement_id:
        arrangement_id = os.path.splitext(os.path.basename(inpath))[0]

    df = pd.DataFrame(labels)
    df["arrangement_id"] = arrangement_id
    df.to_csv(outpath, sep="\t", index=False)

    # js_outpath = outpath.replace("csv", "json")
    # df.to_json(js_outpath, orient="records")

    logger.info(f"Labels saved to {outpath}")

    return {inpath: labels}


def extract_progression_from_midi(
    filepath,
    shortest_note=DEFAULT_SHORTEST_NOTE,
    smooth_beat=DEFAULT_SMOOTH_BEAT,
    quantize_beat=DEFAULT_QUANTIZE_BEAT,
    simplified_path=None,
    harman_labels_path=None,
):
    """
    Params
        filepath: str, Path to a midi file
        shortest_note: float, see params for `simplify_harmony()`
        smooth_beat: float, see params for `simplify_harmony()`
        quantize_beat: float, see params for `simplify_harmony()`
        simplified_path: str, if passed writes a midi file with simplified harmony. Useful for debugging.
        simplified_path: str, if passed writes a csv file with the output harman labels. Useful for debugging.

    Returns
        progression: Progression, the extracted progression object
        segment_start_times: list(float): the start times of each chord in seconds
    """

    midi = load_midi_file(filepath)

    simplified = simplify_harmony(midi, shortest_note, smooth_beat, quantize_beat)

    if simplified_path:
        simplified.write(simplified_path)
        print(f"Wrote simplified midi to {simplified_path}")

    harman_labels = label_midi(simplified)

    if harman_labels_path:
        _ = write_labels(harman_labels, filepath, harman_labels_path)

    chords = [i["chord"] for i in harman_labels]
    durations = [i["end_time"] - i["start_time"] for i in harman_labels]

    bpms = [i[0] for i in midi.get_tempo_changes() if i[0] > 0]

    # TODO: support tempo changes in progressions
    return Progression(chords, durations, bpm=bpms[0], name=filepath)



import pytest
from chord_progressions.pitch import (
    MIDI_NOTE_FREQUENCIES,
    MIDI_NOTES,
    create_notes_freqs_table,
    get_freq_from_note,
    get_midi_num_from_note,
    get_midi_nums_list_from_midi_nums_str,
    get_n_overtones_harmonic,
    get_note_from_midi_num,
    get_note_list,
    get_note_name_from_note,
    get_octave_from_note,
    get_pitch_class_from_note,
)

EXPECTATIONS = [
    {
        "note": "A#5",
        "name": "A#",
        "octave": 5,
        "midi_num": 82,
        "pitch_class": 10,
        "freq": 932.33,
    },
    {
        "note": "A4",
        "name": "A",
        "octave": 4,
        "midi_num": 69,
        "pitch_class": 9,
        "freq": 440.0,
    },
    {
        "note": "C-1",
        "name": "C",
        "octave": -1,
        "midi_num": 0,
        "pitch_class": 0,
        "freq": 8.18,
    },
    {
        "note": "G9",
        "name": "G",
        "octave": 9,
        "midi_num": 127,
        "pitch_class": 7,
        "freq": 12543.85,
    },
]


def test_get_note_list():

    start = 21
    end = 108

    notes = get_note_list(start, end)

    assert len(notes) == 88


def test_get_note_name_from_note():

    for ex in EXPECTATIONS:
        assert get_note_name_from_note(ex["note"]) == ex["name"]


def test_get_midi_num_from_note():

    for ex in EXPECTATIONS:
        assert get_midi_num_from_note(ex["note"]) == ex["midi_num"]

    with pytest.raises(ValueError):
        get_midi_num_from_note("L5")


def test_get_pitch_class_from_note():

    for ex in EXPECTATIONS:
        assert get_pitch_class_from_note(ex["note"]) == ex["pitch_class"]


def test_get_octave_from_note():
    for ex in EXPECTATIONS:
        assert get_octave_from_note(ex["note"]) == ex["octave"]


def test_get_note_from_midi_num():
    for ex in EXPECTATIONS:
        assert get_note_from_midi_num(ex["midi_num"]) == ex["note"]


def test_get_freq_from_note():
    for ex in EXPECTATIONS:
        assert get_freq_from_note(ex["note"]) == ex["freq"]


def test_create_notes_freqs_table():
    note_freqs = create_notes_freqs_table()

    assert list(note_freqs) == MIDI_NOTES
    assert list(note_freqs.values()) == MIDI_NOTE_FREQUENCIES


def test_get_midi_nums_list_from_midi_nums_str():
    expectations = [
        {"midi_nums_str": "60-48_62-50", "midi_nums_list": [[60, 48], [62, 50]]},
        {"midi_nums_str": "60-48_", "midi_nums_list": [[60, 48]]},
        {"midi_nums_str": "60-48_a", "midi_nums_list": [[60, 48]]},
    ]
    for ex in expectations:
        assert (
            get_midi_nums_list_from_midi_nums_str(ex["midi_nums_str"])
            == ex["midi_nums_list"]
        )


def test_get_n_overtones_harmonic():
    assert get_n_overtones_harmonic(440, 4) == [440, 880, 1320, 1760, 2200]



"""Process polyphonic midi to make it more well-suited for chord extraction"""

from functools import partial

import numpy as np
import pandas as pd
import pretty_midi
from chord_progressions.io.midi import load_midi_file

# note values, e.g. 1/4 = quarter note, 1/64 = 64th note
DEFAULT_SHORTEST_NOTE = 1 / 64
DEFAULT_SMOOTH_BEAT = 1
DEFAULT_QUANTIZE_BEAT = 1 / 2

def determine_key_from_file(midi):
    # TODO integrate this into lib
    """import music21

    filename = 'filename.mid'

    s = music21.converter.parse(filename)

    kd1 = music21.analysis.discrete.KrumhanslSchmuckler()
    kd2 = music21.analysis.discrete.TemperleyKostkaPayne()
    kd3 = music21.analysis.discrete.BellmanBudge()

    key1 = kd1.getSolution(s).name
    key2 = kd2.getSolution(s).name
    key3 = kd3.getSolution(s).name

    print(f"KrumhanslSchmuckler: {key1}\nTemperleyKostkaPayne: {key2}\nBellmanBudge: {key3}")"""

def parse_notes(pmid):
    """Parses a pretty_midi object into a 2D array with (start, end, note) columns."""
    points = []

    for instrument in pmid.instruments:
        if instrument.is_drum:
            continue

        for note in instrument.notes:
            points.append((note.start, note.end, note.pitch))

    dtype = np.dtype([("start", float), ("end", float), ("note", int)])
    notes = np.array(points, dtype=dtype)

    return np.sort(notes, order="start")


def round_to_target(target, x):
    """Round `x` to the closest `target`

    e.g.
        round_to_target(1.12, .5) == 1.00
        round_to_target(1.12, .25) == 1.00
        round_to_target(1.13, .25) == 1.25
    """
    y = 1 / target
    return round(x * y) / y


def quantize_times(times, beat):
    """Quantize an array of times to the closest beat.

    e.g. quantize_times([1.1, 1.2, 1.3, 1.4], 0.5) == [1., 1., 1.5, 1.5]
    """
    func = partial(round_to_target, beat)

    return list(map(func, times))


def drop_short_notes(notes, shortest_duration):
    """Removes notes that are shorter than shortest_duration"""

    short_note_ixs = []

    for ix, n in enumerate(notes):
        if n["end"] - n["start"] < shortest_duration:
            short_note_ixs.append(ix)

    return np.delete(notes, short_note_ixs)


def smooth_notes(notes, step_size):
    """Joins adjacent, repeated notes into single, longer notes"""

    df = pd.DataFrame(notes)

    steps = np.arange(df.start.min(), df.end.max() + step_size, step_size)

    for ix, s in enumerate(steps):
        if ix == 0:
            continue

        seg = df[(df["start"] >= steps[ix - 1]) & (df["end"] < s)]

        vcs = seg.note.value_counts()
        vcs1 = vcs[vcs > 1]

        for k, v in vcs1.items():
            svc = seg[seg.note == k]

            df.loc[svc.index[0], "end"] = seg.loc[svc.index[-1]]["end"]

            df.drop(index=svc.index[1:], inplace=True)

    tups = list(df.itertuples(index=False, name=None))

    dtype = np.dtype([("start", float), ("end", float), ("note", int)])

    return np.array(tups, dtype=dtype)


def mk_midi_from_notes(notes):
    midi = pretty_midi.PrettyMIDI()
    piano_program = pretty_midi.instrument_name_to_program("Acoustic Grand Piano")
    piano = pretty_midi.Instrument(program=piano_program)

    for e in notes:
        note = pretty_midi.Note(
            velocity=100, pitch=e["note"], start=e["start"], end=e["end"]
        )
        piano.notes.append(note)

    midi.instruments.append(piano)

    return midi


def simplify_harmony_from_file(
    filepath,
    shortest_note=DEFAULT_SHORTEST_NOTE,
    smooth_beat=DEFAULT_SMOOTH_BEAT,
    quantize_beat=DEFAULT_QUANTIZE_BEAT,
):
    midi = load_midi_file(filepath)
    return simplify_harmony(midi, shortest_note, smooth_beat, quantize_beat)


def simplify_harmony(
    midi,
    shortest_note=DEFAULT_SHORTEST_NOTE,
    smooth_beat=DEFAULT_SMOOTH_BEAT,
    quantize_beat=DEFAULT_QUANTIZE_BEAT,
):
    """Simplify midi to its essential harmonic content"""

    notes = parse_notes(midi)

    tempos = midi.get_tempo_changes()

    # TODO: handle tempo changes
    if len(tempos[0]) == 0:
        raise ValueError("No tempos")
    if len(tempos[0]) > 1:
        #     raise ValueError("Can't handle tempo changes")
        print(
            "Warning: MIDI file has tempo changes but a constant BPM is assumed when processing"
        )

    # TODO: calculate note durations based on time signature
    bpm = tempos[1][0]
    whole_note_dur = 60 / bpm * 4  # seconds

    # convert durations to seconds
    shortest_dur = whole_note_dur * shortest_note
    smooth_dur = whole_note_dur * smooth_beat
    quantize_dur = whole_note_dur * quantize_beat

    notes_cleaned = drop_short_notes(notes, shortest_dur)
    notes_smoothed = smooth_notes(notes_cleaned, smooth_dur)
    notes_cleaned_2 = drop_short_notes(notes_smoothed, shortest_dur * 4)

    # TODO: split on long rests (parameterized duration)
    # TODO: lengthen notes to fill in shorter rests

    # Quantize both the start and end times of each note
    notes_cleaned_2["start"] = quantize_times(notes_cleaned_2["start"], quantize_dur)
    notes_cleaned_2["end"] = quantize_times(notes_cleaned_2["end"], quantize_dur)

    # If start and end were quantized to the same time, move the end time back
    for note in notes_cleaned_2:
        if note["start"] == note["end"]:
            note["end"] += quantize_dur

    return mk_midi_from_notes(notes_cleaned_2)
